local _IsTable = function(thing)
	local t = type(thing)
	return (t == "userdata" or t == "table")
end

local _IsFragment = function(thing)
	return _IsTable(thing) and thing.SetConditional ~= nil and type(thing.SetConditional) == "function"
end

local _IsControl = function(thing)
	return _IsTable(thing) and thing.SetHidden ~= nil and type(thing.SetHidden) == "function"
end

local _TryForFragment = function(control)
	for _,scene in ipairs{ HUD_SCENE, HUD_UI_SCENE } do
		for _,fragment in ipairs(scene.fragments) do
			if fragment.control == control then return fragment end
		end
	end
	return control
end

local _TryForControl = function(fragment)
	return fragment.control or fragment
end

----------------------------------------------------------------------------------------------------------------
local fragment_scenes = {}
local _AddToAllScenes = function(fragment)
	local AllScenes = SCENE_MANAGER.scenes
	if not fragment_scenes[fragment] then fragment_scenes[fragment] = {} end

	for name,scene in pairs(AllScenes) do
		if not scene:HasFragment(fragment) then
			fragment_scenes[fragment][name] = true
		end
		scene:AddFragment(fragment)
	end
end

local _RestoreFragment = function(fragment)
	local AllScenes = SCENE_MANAGER.scenes
	if not fragment_scenes[fragment] then return end

	for name,scene in pairs(AllScenes) do
		if not fragment_scenes[fragment][name] and scene:HasFragment(fragment) then
			scene:RemoveFragment(fragment)
		end
	end
	fragment_scenes[fragment] = nil
end
----------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------
--Generic functions for typical elements
local generic_enableFunc = function(self)
	assert(self ~= nil and type(self) == "table", "generic_enableFunc: self is invalid.")
	if self.enabled == true then return end

	for _,element in ipairs(self.elements) do
		if _IsFragment(element) then
			element.oldConditional = element.conditional
			if self.combineConditionals then
				element:SetConditional(function() return element.oldConditional() and Immers._EvaluateConditions(self.name) end)
			else
				element:SetConditional(function() return Immers._EvaluateConditions(self.name) end)
			end
			_AddToAllScenes(element)
		elseif _IsControl(element) then
			element:SetHandler("OnShow", function() element:SetHidden(self.shouldBeHidden) end)
			element:SetHandler("OnHide", function() element:SetHidden(self.shouldBeHidden) end)
		end
	end

	self.enabled = true
end

local generic_disableFunc = function(self)
	assert(self ~= nil and type(self) == "table", "generic_disableFunc: self is invalid.")
	if self.enabled == false then return end

	for _,element in pairs(self.elements) do
		if _IsFragment(element) then
			_RestoreFragment(element)
			element:SetConditional(element.oldConditional)
		elseif _IsControl(element) then
			element:SetHandler("OnShow", nil)
			element:SetHandler("OnHide", nil)
		end
	end

	self.enabled = false
end

local generic_updateFunc = function(self)
	self.shouldBeHidden = not Immers._EvaluateConditions(self.name)
	--if self.updateConditional ~= nil then self.shouldBeHidden = (self.updateConditional() and self.shouldBeHidden) end

	for _,element in pairs(self.elements) do
		if _IsFragment(element) then
			--element:Refresh()
			if self.shouldBeHidden ~= element:IsHidden() then element:Refresh() end
		elseif _IsControl(element) then
			element:SetHidden(self.shouldBeHidden)
		end
	end
end

local _ElementFromTemplate = function(elementName, data)
	if #data > 0 then return {
		name = elementName,
		elements = data,
		shouldBeHidden = false,
		enabled = false,
		enableFunc = generic_enableFunc,
		disableFunc = generic_disableFunc,
		updateFunc = generic_updateFunc,
	}
	else
		return nil
	end
end
Immers._ElementFromTemplate = _ElementFromTemplate
----------------------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------------------
Immers._LoadElements = function() return {
	_ElementFromTemplate("Experience Bar", {
		PLAYER_PROGRESS_BAR_FRAGMENT,
		PLAYER_PROGRESS_BAR_CURRENT_FRAGMENT,
	}),
	_ElementFromTemplate("Compass", { 
		_TryForControl(COMPASS_FRAME_FRAGMENT),
	}),
	_ElementFromTemplate("Reticle", { 
		ZO_ReticleContainer
	}),
	_ElementFromTemplate("Skill Bar", { 
		_TryForControl(ACTION_BAR_FRAGMENT),
	}),
	-- _ElementFromTemplate("HP/Mag/Stam", { 
	--  PLAYER_ATTRIBUTE_BARS_FRAGMENT,
	--  --ZO_PlayerAttribute,
	--  --PLAYER_ATTRIBUTE_BARS,
	-- }),
	_ElementFromTemplate("Quest Tracker", { 
		FOCUSED_QUEST_TRACKER_FRAGMENT
	}),
	_ElementFromTemplate("Chat Window", { 
		ZO_ChatWindow
	}),
	_ElementFromTemplate("Chat Window Minbar", { 
		ZO_ChatWindowMinBar
	}),
	_ElementFromTemplate("Performance Meters", { 
		PERFORMANCE_METER_FRAGMENT
	}),
	_ElementFromTemplate("AwesomeInfo", { 
		AwesomeInfo
	}),
	_ElementFromTemplate("AUI Minimap", { 
		_TryForFragment(AUI_Minimap_MainWindow) -- will flicker if we use the control over the fragment
	}),
	_ElementFromTemplate("AUI Buffs", { 
		AUI_Buff
	}),
	_ElementFromTemplate("Ravalox'QuestTracker", { 
		_TryForFragment(QuestTrackerWin)
	}),
	_ElementFromTemplate("Fyrakin's Minimap", { 
		Fyr_MM,
		Fyr_MM_Menu,
		Fyr_MM_Coordinates,
		Fyr_MM_ZoneFrame,
	}),
	_ElementFromTemplate("Wykkyd Toolbar", { 
		wykkydsToolbar
	}),
	WAYPOINTIT and (function()
			local e = _ElementFromTemplate("Wykkyd Toolbar", { 
				WAYPOINTIT.FRAGMENT_DIRECTION_ARROW,
				WAYPOINTIT.FRAGMENT_GROUP_DIRECTION_ARROW
			})
			e.combineConditionals = true
			return e
		end)(),
	AUI and {
		name = "AUI Target Attributes",
		shouldBeHidden = false,
		enabled = false,
		hijacked_func = {},
		enableFunc = function(self)
			assert(self ~= nil and type(self) == "table", "AUI Target Attributes.enableFunc: self is invalid.")
			if self.enabled == true then return end

			-- hijack their OnChanged function manually, since prehook wont work here
			self.hijacked_func = AUI.Attributes.Target.OnChanged
			AUI.Attributes.Target.OnChanged = function(...)
				self.hijacked_func(...)
				self:updateFunc()
			end

			self.enabled = true
		end,
		disableFunc = function(self)
			assert(self ~= nil and type(self) == "table", "AUI Target Attributes.disableFunc: self is invalid.")
			if self.enabled == false then return end
			-- restore their function
			AUI.Attributes.Target.OnChanged = self.hijacked_func

			self.enabled = false
		end,
		updateFunc = function(self)
			self.shouldBeHidden = not Immers._EvaluateConditions(self.name)
			--if self.updateConditional ~= nil then self.shouldBeHidden = (self.updateConditional() and self.shouldBeHidden) end

			if self.shouldBeHidden then
				AUI.Attributes.HideFrame(AUI_ATTRIBUTE_TYPE_PRIMARY_TARGET_HEALTH)
			else
				AUI.Attributes.ShowFrame(AUI_ATTRIBUTE_TYPE_PRIMARY_TARGET_HEALTH)
			end
		end,
	},
	AUI and {
		name = "AUI Attributes",
		elements = {
			AUI_ATTRIBUTE_TYPE_PLAYER_HEALTH,
			AUI_ATTRIBUTE_TYPE_PLAYER_MAGICKA,
			AUI_ATTRIBUTE_TYPE_PLAYER_STAMINA,
			AUI_ATTRIBUTE_TYPE_PLAYER_MOUNT,
			--AUI_ATTRIBUTE_TYPE_PLAYER_WEREWOLF,
			--AUI_ATTRIBUTE_TYPE_PLAYER_SIEGE,
			AUI_ATTRIBUTE_TYPE_PLAYER_SHIELD,
		},
		shouldBeHidden = false,
		enabled = false,
		hijacked_func = {},
		enableFunc = function(self)
			assert(self ~= nil and type(self) == "table", "AUI Target Attributes.enableFunc: self is invalid.")
			if self.enabled == true then return end

			-- hijack their OnPowerUpdate function manually, since prehook wont work here
			self.hijacked_func = AUI.Attributes.OnPowerUpdate
			AUI.Attributes.OnPowerUpdate = function(...)
				self.hijacked_func(...)
				self:updateFunc()
			end

			self.enabled = true
		end,
		disableFunc = function(self)
			assert(self ~= nil and type(self) == "table", "AUI Target Attributes.disableFunc: self is invalid.")
			if self.enabled == false then return end
			-- restore their function
			AUI.Attributes.OnPowerUpdate = self.hijacked_func

			self.enabled = false
		end,
		updateFunc = function(self)
			self.shouldBeHidden = not Immers._EvaluateConditions(self.name)
			--if self.updateConditional ~= nil then self.shouldBeHidden = (self.updateConditional() and self.shouldBeHidden) end

			if self.shouldBeHidden then
				for _,id in ipairs(self.elements) do
					AUI.Attributes.HideFrame(id)
				end
			else
				for _,id in ipairs(self.elements) do
					AUI.Attributes.ShowFrame(id)
				end
			end
		end,
	},
	Harvest and {
		name = 'HarvestMap Pins',
		shouldBeHidden = false,
		enabled = false,

		enableFunc = function(self)
			assert(self ~= nil and type(self) == "table", "AUI Target Attributes.enableFunc: self is invalid.")
			if self.enabled == true then return end

			self.LMP = LibStub("LibMapPins-1.0")
			self.HarvestPins = {}

			--hopefully the user has their HarvestPin settings configured beforehand,
			--	maybe think of a way to push element specific warnings back up to the user for just this sort of case.
			for _,pinType in ipairs(Harvest.PINTYPES) do
				if Harvest.savedVars["settings"].isPinTypeVisible[ Harvest.GetPinType( pinType ) ] then
					self.HarvestPins[#self.HarvestPins+1] = pinType
				end
			end

			self.enabled = true
		end,

		disableFunc = function(self)
			assert(self ~= nil and type(self) == "table", "AUI Target Attributes.disableFunc: self is invalid.")
			if self.enabled == false then return end
			
			for _,pinType in ipairs(self.HarvestPins) do
				self.LMP:SetEnabled(Harvest.GetPinType( pinType ), true)
			end

			self.LMP = nil
			self.HarvestPins = nil

			self.enabled = false
		end,

		updateFunc = function(self)
			local old_shouldBeHidden = self.shouldBeHidden
			self.shouldBeHidden = not Immers._EvaluateConditions(self.name)
			if self.shouldBeHidden == old_shouldBeHidden then return end --if we dont do this, the pins flicker rapidly.

			for _,pinType in ipairs(self.HarvestPins) do
				Harvest.SetPinTypeVisible(pinType, not self.shouldBeHidden)
			end
		end,
	},
	_ElementFromTemplate("SimpleXPBar", { 
		SimpleXPBar and SimpleXPBar.SIMPLEXPBAR_FRAGMENT
	}),

} end

-- local priorState = false
-- function ToggleAllCustomPins()
-- 	priorState = not priorState
-- 	for i=1, COMPASS.container:GetNumChildren() do
-- 		local child = COMPASS.container:GetChild(i)
-- 		if child then child:SetHidden(priorState) end
-- 	end
-- end